import torch
import triton
import triton.language as tl

from .quant_per_block import per_block_int8
from .quant_per_block_hd96 import per_block_int8_hd96

from .attn_qk_int8_per_block_h64 import forward as attn_h64_false
from .attn_qk_int8_per_block_hd128 import forward as attn_h128_false
from .attn_qk_int8_per_block_hd64_causal import forward as attn_h64_true
from .attn_qk_int8_per_block_hd128_causal import forward as attn_h128_true
from .attn_qk_int8_per_block_h96 import forward as attn_h96_false
from .attn_qk_int8_per_block_h96_causal import forward as attn_h96_true

from .attn_qk_int8_per_block_h64_bf16 import forward as attn_h64_false_bf16
from .attn_qk_int8_per_block_hd128_bf16 import forward as attn_h128_false_bf16
from .attn_qk_int8_per_block_hd64_causal_bf16 import forward as attn_h64_true_bf16
from .attn_qk_int8_per_block_hd128_causal_bf16 import forward as attn_h128_true_bf16
from .attn_qk_int8_per_block_h96_bf16 import forward as attn_h96_false_bf16
from .attn_qk_int8_per_block_h96_causal_bf16 import forward as attn_h96_true_bf16


def sageattn(q, k, v, is_causal=False, smooth_k=True):
    q, k, v = q.contiguous(), k.contiguous(), v.contiguous()
    if smooth_k:
        k -= k.mean(dim=-2, keepdim=True)
    headdim = q.size(-1)
    dtype = q.dtype


    if dtype == torch.float16:

        if headdim==96:
            q_int8, q_scale, k_int8, k_scale = per_block_int8_hd96(q, k)
        else:
            q_int8, q_scale, k_int8, k_scale = per_block_int8(q, k)
        
        if is_causal==False:
            if headdim==64:
                o = attn_h64_false(q_int8, k_int8, v, q_scale, k_scale)

            if headdim==96:
                o = attn_h96_false(q_int8, k_int8, v, q_scale, k_scale)

            elif headdim==128:
                o = attn_h128_false(q_int8, k_int8, v, q_scale, k_scale)

        elif is_causal==True:
            if headdim==64:
                o = attn_h64_true(q_int8, k_int8, v, q_scale, k_scale)
            
            if headdim==96:
                o = attn_h96_true(q_int8, k_int8, v, q_scale, k_scale)

            elif headdim==128:
                o = attn_h128_true(q_int8, k_int8, v, q_scale, k_scale)


    elif dtype == torch.bfloat16:

        if headdim==96:
            q_int8, q_scale, k_int8, k_scale = per_block_int8_hd96(q, k)
        else:
            q_int8, q_scale, k_int8, k_scale = per_block_int8(q, k)
        
        if is_causal==False:
            if headdim==64:
                o = attn_h64_false_bf16(q_int8, k_int8, v, q_scale, k_scale)

            if headdim==96:
                o = attn_h96_false_bf16(q_int8, k_int8, v, q_scale, k_scale)

            elif headdim==128:
                o = attn_h128_false_bf16(q_int8, k_int8, v, q_scale, k_scale)

        elif is_causal==True:
            if headdim==64:
                o = attn_h64_true_bf16(q_int8, k_int8, v, q_scale, k_scale)
            
            if headdim==96:
                o = attn_h96_true_bf16(q_int8, k_int8, v, q_scale, k_scale)

            elif headdim==128:
                o = attn_h128_true_bf16(q_int8, k_int8, v, q_scale, k_scale)


    return o



import torch, math
import triton
import triton.language as tl

@triton.jit
def _attn_fwd_inner(acc, l_i, m_i, q, q_scale, 
                    K_ptrs, K_scale_ptr, V_ptrs,  
                    start_m,  
                    BLOCK_M: tl.constexpr, HEAD_DIM: tl.constexpr, BLOCK_N: tl.constexpr,  
                    STAGE: tl.constexpr, offs_m: tl.constexpr, offs_n: tl.constexpr,  
                    N_CTX: tl.constexpr):
    lo, hi = 0, N_CTX
    for start_n in range(lo, hi, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)
        k_mask = offs_n[None, :] < (N_CTX - start_n)   
        k = tl.load(K_ptrs, mask = k_mask)
        k_scale = tl.load(K_scale_ptr)
        qk = tl.dot(q, k).to(tl.float32) * q_scale * k_scale 
        m_ij = tl.maximum(m_i, tl.max(qk, 1))
        qk = qk - m_ij[:, None]
        p = tl.math.exp2(qk)
        l_ij = tl.sum(p, 1)
        
        alpha = tl.math.exp2(m_i - m_ij)
        l_i = l_i * alpha + l_ij
        
        acc = acc * alpha[:, None]
        
        v = tl.load(V_ptrs, mask = offs_n[:, None] < (N_CTX - start_n))
        p = p.to(tl.float16)
        
        acc += tl.dot(p, v.to(tl.float16), out_dtype=tl.float16)     # 
        m_i = m_ij
        K_ptrs += BLOCK_N * HEAD_DIM
        K_scale_ptr += 1
        V_ptrs += BLOCK_N * HEAD_DIM
    return acc, l_i

@triton.jit
def _attn_fwd(Q, K, V, Q_scale, K_scale, Out,  
              stride_qz, stride_qh, stride_qm, stride_qk,  
              stride_kz, stride_kh, stride_kn, stride_kk,  
              stride_vz, stride_vh, stride_vk, stride_vn,  
              stride_oz, stride_oh, stride_om, stride_on,  
              Z, H, N_CTX,  
              HEAD_DIM: tl.constexpr,  
              BLOCK_M: tl.constexpr,  
              BLOCK_N: tl.constexpr,  
              STAGE: tl.constexpr  
              ):
    start_m = tl.program_id(0)
    off_hz = tl.program_id(1)
    
    off_z = off_hz // H
    off_h = off_hz % H
    qvk_offset = off_z.to(tl.int64) * stride_qz + off_h.to(tl.int64) * stride_qh
    vk_offset = qvk_offset // stride_qm
    q_scale_offset = off_hz * tl.cdiv(N_CTX, BLOCK_M)
    k_scale_offset = off_hz * tl.cdiv(N_CTX, BLOCK_N)  
    
    
    
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_n = tl.arange(0, BLOCK_N)
    offs_k = tl.arange(0, HEAD_DIM)
    Q_ptrs = Q + qvk_offset + offs_m[:, None] * stride_qm + offs_k[None, :] * stride_qk
    Q_scale_ptr = Q_scale + q_scale_offset + start_m
    K_ptrs = K + qvk_offset + offs_k[:, None] + offs_n[None, :] * stride_kn
    K_scale_ptr = K_scale + k_scale_offset
    V_ptrs = V + qvk_offset + offs_n[:, None] * stride_qm + offs_k[None, :] * stride_qk
    O_block_ptr = Out + qvk_offset + offs_m[:, None] * stride_qm + offs_k[None, :] * stride_qk
    
    m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    l_i = tl.zeros([BLOCK_M], dtype=tl.float32) + 1.0
    acc = tl.zeros([BLOCK_M, HEAD_DIM], dtype=tl.float32)
    
    
    q = tl.load(Q_ptrs, mask = offs_m[:, None] < N_CTX)
    q_scale = tl.load(Q_scale_ptr)
    acc, l_i = _attn_fwd_inner(acc, l_i, m_i, q, q_scale, K_ptrs, K_scale_ptr, V_ptrs,  
                                    start_m,  
                                    BLOCK_M, HEAD_DIM, BLOCK_N,  
                                    4 - STAGE, offs_m, offs_n, N_CTX 
                                    )
    acc = acc / l_i[:, None]
    tl.store(O_block_ptr, acc.to(Out.type.element_ty), mask = (offs_m[:, None] < N_CTX))


def forward(q, k, v, q_scale, k_scale):
    BLOCK_M = 128
    BLOCK_N = 64
    HEAD_DIM_K = k.shape[-1]
    o = torch.empty_like(q, dtype=torch.bfloat16)
    stage = 1

    grid = (triton.cdiv(q.shape[2], BLOCK_M), q.shape[0] * q.shape[1], 1)
    _attn_fwd[grid](
        q, k, v, q_scale, k_scale, o,  
        q.stride(0), q.stride(1), q.stride(2), q.stride(3),  
        k.stride(0), k.stride(1), k.stride(2), k.stride(3),  
        v.stride(0), v.stride(1), v.stride(2), v.stride(3),  
        o.stride(0), o.stride(1), o.stride(2), o.stride(3),  
        q.shape[0], q.shape[1],  
        N_CTX=q.shape[2],  
        BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N, HEAD_DIM=HEAD_DIM_K,  
        STAGE=stage,  
        num_warps=4,  
        num_stages=3)
    return o




from setuptools import setup, find_packages

setup(
    name='sageattention', 
    version='1.0.0',  
    author='Jintao Zhang',  
    author_email='jt-zhang6@gmail.com', 
    packages=find_packages(),  
    description='Accurate and efficient 8-bit plug-and-play attention.',  
    long_description=open('README.md').read(),  
    long_description_content_type='text/markdown', 
    url='https://github.com/jt-zhang/SageAttention', 
    license='MIT',  
    python_requires='>=3.9', 
    classifiers=[  
        'Development Status :: 3 - Alpha', 
        'Intended Audience :: Developers',  
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3', 
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
    ],
)
